// for logging display, 
// https://knowm.org/how-to-make-real-time-charts-in-java/
/******************************************************
Arduino I2C Sniffer

2011 J. M. De Cristofaro
CC-BY-SA 3.0
this code is unsupported -- use at your own risk!

connect one digital pin each to SCL and SDA
use pins on PORTC (Arduino UNO/Duemilanove pins #0 - 7)
do not connect to serial pins (0 & 1)!

connect GND on Arduino to GND of I2C bus
*******************************************************/

/** DESCRIPTION **/
/*******************************************************
this code runs a "one-shot" capture, meaning that
it only captures data once and then dumps it to the
serial port.

it starts capturing when it detects the SDA line has
gone low. it does not check the SCL line (as a proper
start condition detector would). 

you can adjust the capture window to suit you needs.
a value of 300 is long enough to catch one byte sent
over i2c at the 100 kbit/s standard rate. the sample
data is good enough to show you the sequence of events,
but little else.

if you need very specific, time-aligned data, you should
use a logic analyzer or sampling oscilloscope.
*******************************************************/
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdint.h>
#include <avr/pgmspace.h>
#include "la.h"

#if 0
#define EEDEVADR 0b10100000
#define EEPAGES 128
#define EEPAGESIZE 16
#define EEMAXADDR 0x0800
#define ERROR 1
#define SUCCESS (!ERROR)
#endif


#define byte uint8_t
#define word uint16_t
#define dword uint32_t
#define qword uint64_t

#ifndef F_CPU
#define F_CPU 16000000UL
#endif

#ifndef BAUD
#define BAUD 9600
#endif
#include <util/setbaud.h>
#include <util/parity.h>

void uart_init(void) {
    UBRR0H = UBRRH_VALUE;
    UBRR0L = UBRRL_VALUE;
    
#if USE_2X
    UCSR0A |= _BV(U2X0);
#else
    UCSR0A &= ~(_BV(U2X0));
#endif

    UCSR0C = _BV(UCSZ01) | _BV(UCSZ00); /* 8-bit data */ 
    UCSR0B = _BV(RXEN0) | _BV(TXEN0);   /* Enable RX and TX */    
}

void putch(char c) {
    if (c == '\n') {
        putch('\r');
    }
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = c;
}

char kbhit() {
    return bit_is_set(UCSR0A, RXC0);
}

char getch() {
    loop_until_bit_is_set(UCSR0A, RXC0);
    return UDR0;
}

static void printlong(unsigned short num) { unsigned short tmp;
  if ((tmp=num / (unsigned short)10) != 0) printlong(tmp);
  putch((char)(num % (unsigned short)10) + '0');
}
//----------------------------------------------------------------------------------------------------------
// Print a string
void printstring(char *buf)
{
  while(*buf) putch(*buf++);
}


/************************************************************************************************************/
/* code */
/************************************************************************************************************/
void TWIInit(void)
{
    //set SCL to 400kHz
    TWSR = 0x00;
    TWBR = 0x0C;
    //enable TWI
    TWCR = (1<<TWEN);
}
void TWIStart(void)
{
    TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
}

//send stop signal
void TWIStop(void)
{
    TWCR = (1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
}

void TWIWrite(uint8_t u8data)
{
    TWDR = u8data;
    TWCR = (1<<TWINT)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
}

uint8_t TWIReadACK(void)
{
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
    while ((TWCR & (1<<TWINT)) == 0);
    return TWDR;
}

uint8_t TWIReadNACK(void)
{
    TWCR = (1<<TWINT)|(1<<TWEN);
    while ((TWCR & (1<<TWINT)) == 0);
    return TWDR;
}

uint8_t TWIGetStatus(void)
{
    uint8_t status;
    //mask status
    status = TWSR & 0xF8;
    return status;
}

uint8_t EEWriteByte(uint16_t u16addr, uint8_t u8data)
{
    TWIStart();
    if (TWIGetStatus() != 0x08)
        return ERROR;
    //select devise and send A2 A1 A0 address bits
    TWIWrite((EEDEVADR)|(uint8_t)((u16addr & 0x0700)>>7));
    if (TWIGetStatus() != 0x18)
        return ERROR;  
    //send the rest of address
    TWIWrite((uint8_t)(u16addr));
    if (TWIGetStatus() != 0x28)
        return ERROR;
    //write byte to eeprom
    TWIWrite(u8data);
    if (TWIGetStatus() != 0x28)
        return ERROR;
    TWIStop();
    return SUCCESS;
}

uint8_t EEReadByte(uint16_t u16addr, uint8_t *u8data)
{
    //uint8_t databyte;
    TWIStart();
    if (TWIGetStatus() != 0x08)
        return ERROR;
    //select devise and send A2 A1 A0 address bits
    TWIWrite((EEDEVADR)|((uint8_t)((u16addr & 0x0700)>>7)));
    if (TWIGetStatus() != 0x18)
        return ERROR;
    //send the rest of address
    TWIWrite((uint8_t)(u16addr));
    if (TWIGetStatus() != 0x28)
        return ERROR;
    //send start
    TWIStart();
    if (TWIGetStatus() != 0x10)
        return ERROR;
    //select devise and send read bit
    TWIWrite((EEDEVADR)|((uint8_t)((u16addr & 0x0700)>>7))|1);
    if (TWIGetStatus() != 0x40)
        return ERROR;
    *u8data = TWIReadNACK();
    if (TWIGetStatus() != 0x58)
        return ERROR;
    TWIStop();
    return SUCCESS;
}

uint8_t EEWritePage(uint8_t page, uint8_t *u8data)
{
    //calculate page address
    uint8_t u8paddr = 0;
    uint8_t i;
    u8paddr = page<<4;
    TWIStart();
    if (TWIGetStatus() != 0x08)
        return ERROR;
    //select page start address and send A2 A1 A0 bits send write command
    TWIWrite(((EEDEVADR)|(u8paddr>>3))&(~1));
    if (TWIGetStatus() != 0x18)
        return ERROR;
    //send the rest of address
    TWIWrite((u8paddr<<4));
    if (TWIGetStatus() != 0x28)
        return ERROR;
    //write page to eeprom
    for (i=0; i<16; i++)
    {
        TWIWrite(*u8data++);
            if (TWIGetStatus() != 0x28)
                return ERROR;
    }
    TWIStop();
    return SUCCESS;
}
uint8_t EEReadPage(uint8_t page, uint8_t *u8data)
{
    //calculate page address
    uint8_t u8paddr = 0;
    uint8_t i;
    u8paddr = page<<4;
    TWIStart();
    if (TWIGetStatus() != 0x08)
        return ERROR;
    //select page start address and send A2 A1 A0 bits send write command
    TWIWrite(((EEDEVADR)|(u8paddr>>3))&(~1));
    if (TWIGetStatus() != 0x18)
        return ERROR;
    //send the rest of address
    TWIWrite((u8paddr<<4));
    if (TWIGetStatus() != 0x28)
        return ERROR;
    //send start
    TWIStart();
    if (TWIGetStatus() != 0x10)
        return ERROR;
    //select devise and send read bit
    TWIWrite(((EEDEVADR)|(u8paddr>>3))|1);
    if (TWIGetStatus() != 0x40)
        return ERROR;
    for (i=0; i<15; i++)
    {
        *u8data++ = TWIReadACK();
            if (TWIGetStatus() != 0x50)
                return ERROR;
    }  
    *u8data = TWIReadNACK();
    if (TWIGetStatus() != 0x58)
        return ERROR;
    TWIStop();
    return SUCCESS;
}

/************************************************************************************************************/
/* code */
/************************************************************************************************************/
#define NOP __asm__ __volatile__ ("nop\n\t") 

// assign pin values here
const char sda_pin = 2;
const char scl_pin = 7;

// sampling window
#define data_size 1536
// array to contain sampled data
byte captured_data[data_size];

// housekeeping booleans
#define captured 1
#define dumped 	 2
#define backward 4
byte flag;

byte trigger_xor,trigger_and;

void capture_dly(word d);
void capture_dly_(word d);

void setup()
  {
    uart_init();
    DDRC=0;
    //pinMode(sda_pin, INPUT);
    //pinMode(scl_pin, INPUT);
    flag=0;
    printstring("Good to go, chief!");
  }

byte type;
// main loop: waits for SDA to go low, then
// samples the data, then formats and dumps it
// to the serial port.
void loop()
  {  byte tmp; flag=0;
	trigger_xor&=trigger_and;
    while((tmp=PINC)&trigger_and^trigger_xor);
	switch(type) {  default:
		case 0: capture(tmp); break;
		case 1: captureRs232(); break;
		case 2: captureRs232_2(); break;
		case 3: captureRs232_3(); break;
		case 4: captureRs232_4(); break;
		case 5: captureRs232C(tmp); break;
		case 6: captureXor(tmp); break;
		case 7: capture4(); break;
		case 8: capture8(); break;
		case 9: capture_(); break;
		case 10: capture_dly_(13); break;
		case 11: capture_dly(13); break;
	};
    if (flag&captured&& ~flag&dumped)
      {
        serial_dump();
		flag|=dumped;
      }
  }

// captures the data on PORTC and stores it in a global array
void capture8()
  {
#if USE_CAPTURE8
   #include "la8.h"
#endif
    flag|=(captured);
  }

// captures the data on PORTC and stores it in a global array
void capture4()
  { byte j;
#if USE_CAPTURE4
   #define SWAP
   #include "la8.h"
#endif
    flag|=(captured);
  }

// captures the data on PORTC and stores it in a global array
void capture(byte data)
  {
    byte tempdata; int x;
	captured_data[0] = data&63;
    for (x = 1; x < data_size; x++)
      {
        tempdata = PINC;
        captured_data[x] = tempdata;
      }
    flag|=(captured);
  }

// captures the data on PORTC and stores it in a global array
void capture_()
{
  #if 0
asm volatile ( "\n\t"
        "ldi r30,lo8(captured_data+1536) \n\t"
        "ldi r31,hi8(captured_data+1536) \n\t"
        "ldi r24,hi8(-data_size) \n\t"
        "ldi r25,lo8(-data_size) \n"
		"L_loop%=: \n\t"
        "in __tmp_reg__,0x6	\n\t"
        "st -Z,__tmp_reg__	\n\t"
        "adiw r24,1	\n\t"
        "brcs L_done%=\n\t"
        "rjmp L_loop%=\n"
		"L_done%=:	\n\t"
		::: "r31","r30","r24","r25" 
	);
#else
  word i;
  for (i=data_size; i; i--)
    captured_data[i-1] = PINC;
#endif
   flag|=(captured)|backward;
  }

// captures the data on PORTC and stores it in a global array
void capture_dly_(word d)
  {
  word i,t;
  for (i=data_size; i; i--) {
	for(t=d;t--;) NOP;
    captured_data[i-1] = PINC;
   }
   flag|=(captured)|backward;
  }
// captures the data on PORTC and stores it in a global array
void capture_dly(word d)
  {
  word i,t; 
  i=data_size;
  for(i=data_size;i;i--) {
    captured_data[i-1] = PINC;
	NOP;
	if(d&1) continue;
	t=d>>1; do { NOP;  } while(t--); ;
 }  
   flag|=(captured)|backward;
  }

// captures the data on PORTC and stores it in a global array
void captureXor(byte data)
  {
    byte tempdata, last=data&63, cnt; int x;
    for (x = 0; x < data_size-1; x++)
      { cnt=0; 
        tempdata = PINC&0x3f;
		if((last^tempdata)) goto save;					// different sample
        tempdata = PINC&0x3f;
		if((last^tempdata)) { last|=0x40; goto save; }	// one repetition
		do
			tempdata = PINC&0x3f;
		while(++cnt&&!(last^tempdata)); 
		if(cnt==1) { last|=0x80; goto save; }			// two repetitions
		captured_data[x++]=last|0xc0; last=cnt; 
save:	captured_data[x] =last; next: last=tempdata;
     }
		if(x<data_size) captured_data[x]=PINC&0x3f;
    flag|=(captured);
  }

// captures the data on PORTC and stores it in a global array
void captureRs232()
  {
    byte tempdata,tmp; 
	for (;;)
      { 
		if(kbhit()&&getch()==0x18) break;
		tempdata=PINC&63;
		if(10000000/BAUD>33) {
			_delay_us(10000000/BAUD/2-7); if(tempdata^(tmp=PINC&63)) tempdata|=0x40;
			_delay_us(10000000/BAUD/3-7); if(tmp^PINC&63) tempdata|=0x80;
		}
		if(kbhit()&&getch()==0x18) break;
		putch(tempdata);
      }
  }
// captures the data on PORTC and stores it in a global array
void captureRs232_4()
  {
    byte tempdata;
    for (;;)
      { tempdata=PINC&15;
		_delay_us(10000000/BAUD/2-5); tempdata|=PINC<<4;
		if(kbhit()&&getch()==0x18) break;
		putch(tempdata);
		tempdata=0;
      }
  }
// captures the data on PORTC and stores it in a global array
void captureRs232_2()
  {
    byte tempdata=0;
    for (;;)
      { tempdata|=PINC&3;
		if(kbhit()&&getch()==0x18) break;
		putch(tempdata);
		_delay_us(10000000/BAUD/4-7); tempdata|=(PINC&3)<<6;
		_delay_us(10000000/BAUD/4-7); tempdata|=(PINC&3)<<4;
		_delay_us(10000000/BAUD/4-7); tempdata|=(PINC&3)<<2;
		_delay_us(10000000/BAUD/4-7); 
		tempdata=0;
      }
  }

void captureRs232_3()
  {
    byte tempdata=0,tmp;
    for (tmp=8;;tmp<<=1,!tmp&&(tmp=8))
      { 
		tempdata=(PINC&7); if(PINC&tmp) tempdata|=0x80; if(tmp==8) tempdata|=tmp;
		_delay_us(10000000/BAUD/4-7); 
		_delay_us(10000000/BAUD/4-7); tempdata|=(PINC&7)<<4;
		_delay_us(10000000/BAUD/4-7);
		_delay_us(10000000/BAUD/4-7); 
		if(kbhit()&&getch()==0x18) break;
		putch(tempdata); 
      }
  }

#define send()    if(tx!=rx&&!bit_is_set(UCSR0A, UDRE0)) UDR0=captured_data[tx++],tx>=data_size&&(tx=0)

// captures the data on PORTC and stores it in a global array
void captureRs232C(byte data)
  {
    byte tempdata, last=data&63, cnt; int x;
	word rx=0,tx=0;
    for (x = 0; x < data_size-1; x++)
      { cnt=0; 
        tempdata = PINC&0x3f;
		if((last^tempdata)) goto save;					// different sample
		send();
		do
			tempdata = PINC&0x3f;
		while(++cnt&&!(last^tempdata));
		send(); while((tx?tx-1:data_size)==rx) send(),last|=0x80; captured_data[rx++]=(last|0x40); last=cnt; 
save:	send(); while((tx?tx-1:data_size)==rx) send(),last|=0x80; captured_data[rx++]=last; next: last=tempdata;
		send();
     }
		while(tx!=rx) send();
  }


static void data(word x) 
    {
      if (x<10)  { putch('0'); }
      if (x<100) { putch('0'); }
      if (x<1000){ putch('0'); }
      printlong(x);
      printstring(",     ");
      printlong(captured_data[x]);
      putch('\n');
#if 0
      temp = bitRead(captured_data[x], sda_pin);
      if (temp == 0) { Serial.print(0); }
      else { Serial.print(1); }
      Serial.print(",   ");
      temp = bitRead(captured_data[x], scl_pin);
      if (temp == 0) { Serial.println(0); }
      else { Serial.println(1); }
#endif
    }

// reads the data out of the global array, formats it
// and outputs it to the serial port.
void serial_dump()
  {
    byte temp; word x;
    printstring("sample, sda, sck\n");
    if(flag&backward) 
	for(x = data_size; x--;)	data(x);
	else for(x = 0; x < data_size; x++)	data(x);
	flag=0;
  }

void main() { setup(); for(;;) loop(); }
